import pandas as pd


def extract_columns(feature_count_table, name_columns, output_table):
    attribute_df = pd.read_csv(feature_count_table, sep='\t')
    at_dict = attribute_df.Attributes.apply(
        lambda attributes: dict([key_value_pair.split('=') for
                                 key_value_pair in attributes.split(';')]))
    for atr in name_columns:
        attribute_df[atr] = at_dict.apply(lambda at_dic:
                                          at_dic.get(atr))
    attribute_df.to_csv(output_table, index=None, sep='\t')




import numpy as np
import networkx as nx
import pandas as pd
from bokeh.io import show, output_file
from bokeh.models import Range1d, MultiLine, Circle, HoverTool, TapTool, ColumnDataSource, \
    BoxZoomTool, ResetTool, PanTool, WheelZoomTool, OpenURL
from bokeh.models.graphs import from_networkx, NodesAndLinkedEdges
from bokeh.palettes import Spectral4
from bokeh.plotting import figure


def plot_network_graph_rna_protein(feature_count_table, index_table, threshold, max_size, output_plot):
    correlated_table = pd.read_csv(feature_count_table, sep='\t')
    correlated_table.set_index(index_table, inplace=True)
    plot_graph(correlated_table, threshold, max_size, output_plot)


def do_plot_graph(nodes, edges, colors, sizes, description, output_plot):
    G = nx.Graph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)

    hover = HoverTool(tooltips=[
        ("name", "@index")])

    plot = figure(plot_width=900, plot_height=900, x_range=Range1d(-1.1, 1.1), y_range=Range1d(-1.1, 1.1),
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title=output_plot)
    plot.toolbar.logo = None

    plot.title.text = description

    url_protein = "https://www.ncbi.nlm.nih.gov/gene/?term=@index"
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url_protein)

    graph_renderer = from_networkx(G, nx.fruchterman_reingold_layout, scale=1)

    source = ColumnDataSource({'index': nodes, 'fill_color': colors, 'size': sizes})
    graph_renderer.node_renderer.data_source = source
    graph_renderer.node_renderer.glyph = Circle(size="size", fill_color="fill_color", line_width=0,
                                                line_color="fill_color")

    graph_renderer.node_renderer.selection_glyph = Circle(size="size", fill_color=Spectral4[2], line_width=0,
                                                          line_color=Spectral4[1])
    graph_renderer.node_renderer.hover_glyph = Circle(size="size", fill_color=Spectral4[1], line_width=0,
                                                      line_color=Spectral4[1])

    graph_renderer.edge_renderer.glyph = MultiLine(line_color="#CCCCCC", line_alpha=1, line_width=0.2)
    graph_renderer.edge_renderer.selection_glyph = MultiLine(line_color=Spectral4[2], line_width=1)
    graph_renderer.edge_renderer.hover_glyph = MultiLine(line_color=Spectral4[1], line_width=1)

    graph_renderer.selection_policy = NodesAndLinkedEdges()
    graph_renderer.inspection_policy = NodesAndLinkedEdges()

    plot.renderers.append(graph_renderer)

    output_file(output_plot)


def plot_graph(correlated_table, threshold, max_size, output_plot):
    description = "Network graph rna-protein"
    nodes = [*correlated_table.index, *correlated_table.columns]

    edges = []
    for index in correlated_table.index:
        for column in correlated_table.columns:
            weight = correlated_table.loc[index, column]
            if weight >= threshold:
                edges.append((index, column))

    colors = [Spectral4[0]] * len(correlated_table.index) + [Spectral4[3]] * len(correlated_table.columns)

    sizes = []

    for index in correlated_table.index:
        s = correlated_table.loc[index]
        size = np.sqrt(s[s >= threshold].sum())
        sizes.append(size)

    for column in correlated_table.columns:
        s = correlated_table.loc[:, column]
        size = np.sqrt(s[s >= threshold].sum())
        sizes.append(size)

    scaled_sizes = [float(i) / max(sizes) * max_size for i in sizes]

    do_plot_graph(nodes, edges, colors, scaled_sizes, description, output_plot)



import pandas as pd
from scipy.stats import spearmanr
from scipy.stats import pearsonr


def corr_specific_gene_vs_all(feature_count_table,
                              feature_count_start_column,
                              feature_count_end_column,
                              name_column_with_genes_name,
                              name, correlation, output_file):
    table_protein = pd.read_table(feature_count_table)
    matrix = _extract_value_matrix(table_protein,
                                   feature_count_start_column,
                                   feature_count_end_column)
    gene_column = _exctract_gene_column(table_protein,
                                        name_column_with_genes_name)
    table_with_genes = pd.concat([gene_column, matrix], axis=1)
    table_with_genes_new_index = \
        table_with_genes.set_index(name_column_with_genes_name)
    gene_row_specific_gene = \
        _extract_gene_row(table_with_genes_new_index, name)
    if correlation == "Spearman":
        spearman_correlation(gene_row_specific_gene,
                             table_with_genes_new_index, output_file)
    else:
        pearson_correlation(gene_row_specific_gene,
                            table_with_genes_new_index, output_file)


def _extract_gene_row(feature_count_table_df, gene_name):
    return feature_count_table_df.loc[feature_count_table_df.index == gene_name]


def _extract_value_matrix(feature_count_table_df, feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:,
                                       feature_count_start_column:feature_count_end_column]


def _exctract_gene_column(feature_count_table_df, name_column_with_genes_name):
    return feature_count_table_df[[name_column_with_genes_name]]


def pearson_correlation(gene_row_specific_gene, table_with_genes_new_index, output_file):
    rho_list = []
    p_value_list = []
    for index_gene, row_gene in table_with_genes_new_index.iterrows():
        rho = pearsonr(gene_row_specific_gene.iloc[0], row_gene)[0]
        rho_list.append(rho)
        p_value = pearsonr(gene_row_specific_gene.iloc[0], row_gene)[1]
        p_value_list.append(p_value)
    table_with_genes_new = table_with_genes_new_index.reset_index()
    rho_df = pd.DataFrame(rho_list)
    p_df = pd.DataFrame(p_value_list)
    table_with_rho_and_pvalue = pd.concat([table_with_genes_new, rho_df, p_df], axis=1)
    table_with_rho_and_pvalue.columns.values[[-2, -1]] = ['Pearson_coefficient', 'p_value']
    table_with_rho_and_pvalue.to_csv(output_file, sep='\t', index=None)


def spearman_correlation(gene_row_specific_gene, table_with_genes_new_index, output_file):
    rho_list = []
    p_value_list = []
    for index_gene, row_gene in table_with_genes_new_index.iterrows():
        rho = spearmanr(gene_row_specific_gene.iloc[0], row_gene)[0]
        rho_list.append(rho)
        p_value = spearmanr(gene_row_specific_gene.iloc[0], row_gene)[1]
        p_value_list.append(p_value)
    table_with_genes_new = table_with_genes_new_index.reset_index()
    rho_df = pd.DataFrame(rho_list)
    p_df = pd.DataFrame(p_value_list)
    table_with_rho_and_pvalue = pd.concat([table_with_genes_new, rho_df, p_df], axis=1)
    table_with_rho_and_pvalue.columns.values[[-2, -1]] = ['Spearman_coefficient', 'p_value']
    table_with_rho_and_pvalue.to_csv(output_file, sep='\t', index=None)



import pandas as pd


def selecting_specific_features(normalized_table, feature_count_start_column,
                                feature_count_end_column, features, output_file):
    normalized_table_df = pd.read_table(normalized_table)
    features_rows = finding_nc_rnas_rows(normalized_table_df, features)
    attribute_table = _extract_attributes(features_rows, feature_count_start_column)
    nc_rnas_table = _extract_value_specific_feature(features_rows, feature_count_start_column,
                                                    feature_count_end_column)
    pd.concat([attribute_table, nc_rnas_table],
              axis=1).to_csv(output_file, sep='\t', index=None)


def finding_nc_rnas_rows(normalized_table, features):
    return normalized_table.loc[normalized_table['Feature'].isin(features)]


def _extract_value_specific_feature(features_rows, feature_count_start_column, feature_count_end_column):
    return features_rows.iloc[:, feature_count_start_column:feature_count_end_column]


def _extract_attributes(features_rows,
                        feature_count_start_column):
    return features_rows.iloc[:, : int(feature_count_start_column)]



import os
import sys
import pandas as pd


def create_root_folder(project_name):
    """Create the root folder of a new project with the given name.

    Arguments:
    - `project_name`: Name of the project root folder
    """
    if not os.path.exists(project_name):
        os.mkdir(project_name)
    else:
        sys.stderr.write("Cannot create folder \"%s\"! File/folder with "
                         "the same name exists already.\n" % project_name)
        sys.exit(2)


def create_subfolders(root, subfolders):
    """Create required subfolders in the given folder.

    Arguments:
    - `project_name`: Name of the project root folder
    """
    for folder in subfolders:
        if not os.path.exists(root + "/" + folder):
            os.mkdir(root + "/" + folder)


def create_version_file(version_file_path, version):
    with open(version_file_path, "w") as fh:
        fh.write("GRADitude version: %s\n" % version)
        fh.write("Python version: %s\n" % sys.version.replace("\n", " "))
        fh.write("pandas version: %s\n" % pd.__version__)




import pandas as pd
from scipy import stats


def rna_protein_correlation(feature_count_table, feature_count_start_column,
                            feature_count_end_column, protein_table,
                            protein_count_start_column, protein_count_end_column,
                            index_sequencing, index_protein,
                            correlation,
                            output_file):
    sequencing_table = pd.read_csv(feature_count_table, sep="\t")
    protein_table = pd.read_csv(protein_table, sep="\t")
    sequencing_table.set_index(index_sequencing, inplace=True)
    sequencing_table.index = \
        sequencing_table.index + '_' + sequencing_table.groupby(
            level=0).cumcount().astype(str)
    protein_table.set_index(index_protein, inplace=True)
    protein_table.index = \
        protein_table.index + '_' + protein_table.groupby(
            level=0).cumcount().astype(str)
    value_matrix_sequencing = _extract_value_matrix(sequencing_table, feature_count_start_column,
                                                    feature_count_end_column)
    value_matrix_protein = _extract_value_matrix(protein_table, protein_count_start_column,
                                                 protein_count_end_column)
    if correlation == "Spearman":
        correlation_spearman(value_matrix_sequencing, value_matrix_protein, output_file)
    else:
        correlation_pearson(value_matrix_sequencing, value_matrix_protein, output_file)


def _extract_value_matrix(feature_count_table_df, feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column) - 1:int(
        feature_count_end_column) - 1]


def correlation_spearman(value_matrix_sequencing, value_matrix_protein, output_file):
    correlation_dataframe = pd.DataFrame(0, columns=value_matrix_sequencing.index,
                                         index=value_matrix_protein.index)

    for index_gene, row_gene in value_matrix_sequencing.iterrows():

        for index_protein, row_protein in value_matrix_protein.iterrows():
            rho = stats.spearmanr(row_gene, row_protein)[0]
            correlation_dataframe.loc[index_protein, index_gene] = rho

    correlation_dataframe.to_csv(output_file, sep='\t', index_label="Gene")


def correlation_pearson(value_matrix_sequencing, value_matrix_protein, output_file):
    correlation_dataframe = pd.DataFrame(0, columns=value_matrix_sequencing.index,
                                         index=value_matrix_protein.index)

    for index_gene, row_gene in value_matrix_sequencing.iterrows():

        for index_protein, row_protein in value_matrix_protein.iterrows():
            rho = stats.pearsonr(row_gene, row_protein)[0]
            correlation_dataframe.loc[index_protein, index_gene] = rho

    correlation_dataframe.to_csv(output_file, sep='\t', index_label="Gene")



import pandas as pd
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
import numpy as np
import umap
from bokeh.plotting import figure, output_file, save, ColumnDataSource
from bokeh.models import HoverTool, BoxZoomTool, ResetTool, PanTool
from bokeh.models import WheelZoomTool, TapTool, OpenURL
import bokeh.palettes


def dr_proteins(feature_count_table, feature_count_start_column,
                feature_count_end_column, dimension_reduction_algorithm,
                perplexity, n_neighbors, min_dist, output_colored_clusters):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column,
                                         feature_count_end_column)
    if dimension_reduction_algorithm == "t-SNE":
        t_sne_result = perform_t_sne(value_matrix,
                                     perplexity)
        plot_t_sne(feature_count_table_df, t_sne_result,
                   output_colored_clusters)
    elif dimension_reduction_algorithm == "PCA":
        pca_result = perform_pca(value_matrix)
        plot_pca(feature_count_table_df, pca_result,
                 output_colored_clusters)
    elif dimension_reduction_algorithm == "UMAP":
        umap_result = perform_umap(value_matrix, n_neighbors, min_dist)
        plot_umap(feature_count_table_df, umap_result,
                  output_colored_clusters)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:,
                                       feature_count_start_column:feature_count_end_column]


def perform_t_sne(normalized_values, perplexity):
    model = TSNE(n_components=2, random_state=0, perplexity=perplexity)
    np.set_printoptions(suppress=True)
    tsne_result = model.fit_transform(normalized_values)
    return tsne_result


def perform_pca(normalized_values):
    model = PCA(n_components=3, random_state=0)
    np.set_printoptions(suppress=True)
    pca_result = model.fit_transform(normalized_values)
    return pca_result


def perform_umap(normalized_values, n_neighbors, min_dist):
    reducer = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist)
    umap_results = reducer.fit_transform(normalized_values)
    return umap_results


def plot_t_sne(read_counting_table, tsne_result, output_file_colorized_by_clusters):
    read_counting_table["t-SNE-component_1"] = [pos[0] for pos in tsne_result]
    read_counting_table["t-SNE-component_2"] = [pos[1] for pos in tsne_result]

    color_palette = bokeh.palettes.Colorblind[(
        len(read_counting_table["Cluster_label"].unique()))]
    color = read_counting_table["Cluster_label"].apply(
        lambda lable: color_palette[lable])
    label = read_counting_table.apply(
        _label_clustering, axis=1)

    hower_data = dict(
        x=read_counting_table["t-SNE-component_1"],
        y=read_counting_table["t-SNE-component_2"],
        feature=read_counting_table["Protein.names"],
        id=read_counting_table["Protein.IDs"],
        cluster_label=read_counting_table["Cluster_label"],
        color=color,
        label=label)

    hower_data_df = pd.DataFrame.from_dict(hower_data)
    hower_data_df.to_csv("hower_data_df.csv", sep='\t', index=None)
    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Protein.names", "@feature"),
        ("Protein.IDs", "@id")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq t-SNE proteins")

    plot.circle("x", "y", source=source, size=7, alpha=3,
                color='color', legend="label", line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'

    url = "https://www.uniprot.org/uniprot/?query=@feature"
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"

    output_file(output_file_colorized_by_clusters)
    save(plot)


def plot_pca(read_counting_table, pca_result, output_file_colorized_by_clusters):
    read_counting_table["PCA-component_1"] = [pos[0] for pos in pca_result]
    read_counting_table["PCA-component_2"] = [pos[1] for pos in pca_result]

    color_palette = bokeh.palettes.Colorblind[(
        len(read_counting_table["Cluster_label"].unique()))]
    color = read_counting_table["Cluster_label"].apply(
        lambda lable: color_palette[lable])
    label = read_counting_table.apply(
        _label_clustering, axis=1)

    hower_data = dict(
        x=read_counting_table["PCA-component_1"],
        y=read_counting_table["PCA-component_2"],
        feature=read_counting_table["Protein.names"],
        id=read_counting_table["Protein.IDs"],
        cluster_label=read_counting_table["Cluster_label"],
        color=color,
        label=label)

    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Protein.names", "@feature"),
        ("Protein.IDs", "@id")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq PCA proteins")

    plot.circle("x", "y", source=source, size=7, alpha=3,
                color='color', legend="label", line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'

    url = "https://www.uniprot.org/uniprot/?query=@feature"
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Component 1"
    plot.yaxis.axis_label = "Component 2"

    output_file(output_file_colorized_by_clusters)
    save(plot)


def plot_umap(read_counting_table, umap_result, output_file_colorized_by_clusters):
    read_counting_table["UMAP-component_1"] = [pos[0] for pos in umap_result]
    read_counting_table["UMAP-component_2"] = [pos[1] for pos in umap_result]

    color_palette = bokeh.palettes.Colorblind[(
        len(read_counting_table["Cluster_label"].unique()))]
    color = read_counting_table["Cluster_label"].apply(
        lambda lable: color_palette[lable])
    label = read_counting_table.apply(
        _label_clustering, axis=1)

    hower_data = dict(
        x=read_counting_table["UMAP-component_1"],
        y=read_counting_table["UMAP-component_2"],
        feature=read_counting_table["Protein.names"],
        id=read_counting_table["Protein.IDs"],
        cluster_label=read_counting_table["Cluster_label"],
        color=color,
        label=label)

    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Protein.names", "@feature"),
        ("Protein.IDs", "@id")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq UMAP proteins")

    plot.circle("x", "y", source=source, size=7, alpha=3,
                color='color', legend="label", line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'
    url = "https://www.uniprot.org/uniprot/?query=@feature"
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"

    output_file(output_file_colorized_by_clusters)
    save(plot)


def _label_clustering(row):
    __label = {0: "cluster1", 1: "cluster2", 2: "cluster3",
               3: "cluster4", 4: "cluster5", 5: "cluster6",
               6: "cluster7", 7: "cluster8"}[row["Cluster_label"]]
    return __label



import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import DBSCAN


def clustering(feature_count_table, feature_count_start_column, feature_count_end_column,
               number_of_clusters, pseudo_count, clustering_methods,
               epsilon, min_samples, scaling_method,
               output_file):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column, feature_count_end_column)
    attribute_matrix = _extract_attributes(feature_count_table_df,
                                           feature_count_start_column)
    normalized_table = normalize_values(value_matrix, scaling_method, pseudo_count)
    if clustering_methods == 'k-means':
        table_with_clusters = k_means_clustering(normalized_table, number_of_clusters)
    elif clustering_methods == 'hierarchical_clustering':
        table_with_clusters = hierarchical_clustering(normalized_table, number_of_clusters)
    elif clustering_methods == 'DBSCAN':
        table_with_clusters = dbscan_clustering(normalized_table, epsilon, min_samples)
    pd.concat([attribute_matrix, table_with_clusters],
              axis=1).to_csv(output_file, sep='\t', index=None)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, feature_count_start_column:feature_count_end_column]


def _extract_attributes(feature_count_table_df,
                        feature_count_start_column):
    return feature_count_table_df.iloc[:, : int(feature_count_start_column)]


def hierarchical_clustering(values_matrix, number_of_clusters):
    h_clustering = AgglomerativeClustering(n_clusters=number_of_clusters,
                                           affinity="euclidean", linkage="ward")
    h_clustering.fit(values_matrix)
    labels = h_clustering.labels_
    pd.DataFrame(data=labels, columns=['cluster'])
    values_matrix["Cluster_label"] = pd.Series(h_clustering.labels_).astype(int)
    return values_matrix


def k_means_clustering(values_matrix, number_of_clusters):
    k_means = KMeans(n_clusters=number_of_clusters, random_state=0)
    k_means.fit(values_matrix)
    labels = k_means.labels_
    pd.DataFrame(data=labels, columns=['cluster'])
    values_matrix["Cluster_label"] = pd.Series(k_means.labels_).astype(int)
    return values_matrix


def dbscan_clustering(values_matrix, epsilon, min_samples):
    dbscan = DBSCAN(eps=epsilon, min_samples=min_samples)
    dbscan.fit_predict(values_matrix)
    labels = dbscan.labels_
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    print('Estimated number of clusters: %d' % n_clusters_)
    pd.DataFrame(data=labels, columns=['cluster'])
    values_matrix["Cluster_label"] = pd.Series(dbscan.labels_).astype(int)
    return values_matrix


def normalize_values(values_matrix, scaling_method, pseudo_count):
    if scaling_method == "no_normalization":
        normalized_values = values_matrix
    elif scaling_method == "log2":
        normalized_values = values_matrix.applymap(
            lambda val: val + pseudo_count).applymap(np.log2)
    elif scaling_method == "log10":
        normalized_values = values_matrix.applymap(
            lambda val: val + pseudo_count).applymap(np.log10)
    elif scaling_method == "to_max":
        values_matrix = values_matrix.fillna(lambda x: 0)
        row_max_values = values_matrix.max(axis=1)
        normalized_values = values_matrix.divide(
            row_max_values, axis=0)
        normalized_values = pd.DataFrame(normalized_values).fillna(0)
    elif scaling_method == "to_range":
        values_matrix = values_matrix.fillna(lambda x: 0)
        row_max_values = values_matrix.max(axis=1)
        row_min_values = values_matrix.min(axis=1)
        normalized_values = values_matrix.subtract(
            row_min_values, axis=0).divide(row_max_values, axis=0)
        normalized_values = pd.DataFrame(normalized_values).fillna(0)
    else:
        print("Normalization method not known")
    return normalized_values



import pandas as pd
from scipy import stats


def correlation(
        feature_count_table, feature_count_start_column, feature_count_end_column,
        correlation_type,
        output_table):
    feature_count_table_df = pd.read_table(feature_count_table, sep='\t')
    feature_count_table_df.set_index(['Gene'], inplace=True)
    feature_count_table_df_value = _extract_value_matrix(feature_count_table_df,
                                                         feature_count_start_column,
                                                         feature_count_end_column)
    if correlation_type == "Spearman":
        corr_hist_spearman(feature_count_table_df_value, output_table)
    else:
        corr_hist_pearson(feature_count_table_df_value, output_table)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):feature_count_end_column]


def corr_hist_spearman(feature_count_table_df_value, output_table):
    rho = stats.spearmanr(feature_count_table_df_value.values.T)[0]
    rho_df = pd.DataFrame(rho, columns=feature_count_table_df_value.index,
                          index=feature_count_table_df_value.index)
    rho_df.to_csv(output_table, sep="\t")


def corr_hist_pearson(feature_count_table_df_value, output_table):
    rho = stats.spearmanr(feature_count_table_df_value.values.T)
    rho_df = pd.DataFrame.from_records(rho, columns=feature_count_table_df_value.index,
                                       index=feature_count_table_df_value.index)

    rho_df.to_csv(output_table, sep="\t")



import pandas as pd
import numpy as np
from numpy import inf
import matplotlib.pyplot as plt
from sklearn import linear_model
from sklearn.linear_model import RANSACRegressor
plt.switch_backend('agg')


def robust_regression(
        ref_feature_count_table, concentration_table, number_of_outliers, number_of_ercc_in_common, used_mix,
        output_file):
    ref_feature_count_table_df = read_table(ref_feature_count_table)
    concentration_table_df = read_table_concentration(concentration_table)
    gradient_file_combined = merging_dict(ref_feature_count_table_df, concentration_table_df, used_mix)
    read_grad_value, read_concentration_value, gradient_file = \
        modify_input(gradient_file_combined)
    common_ercc_df = regression(read_grad_value, read_concentration_value, gradient_file,
                                number_of_outliers)
    df_common_ercc = common_ercc(common_ercc_df)
    new_ref_table(df_common_ercc, ref_feature_count_table_df, output_file, number_of_ercc_in_common)


def read_table(ref_feature_count_table):
    """Read the spike-in table as a dictionary"""
    ref_feature_count_table_dict = {}
    ref_feature_count_table_df = pd.read_csv(ref_feature_count_table, sep='\t')
    for index, row in ref_feature_count_table_df.iterrows():
        key = row[0][:row[0].index(' ')]
        ref_feature_count_table_dict[key] = row
    return ref_feature_count_table_dict


def read_table_concentration(concentration_table):
    """Read the spike-in concentration table"""
    new_dict_for_conc_table = {}
    concentration_table_df = pd.read_table(concentration_table)
    for index, row in concentration_table_df.iterrows():
        new_dict_for_conc_table[row[1]] = row
    return new_dict_for_conc_table


def merging_dict(ref_feature_count_table_dict, new_dict_for_conc_table, mix):
    """Merge the two dictionaries using common keys. """
    combined_dict = {}
    concentration_list = []
    for key in ref_feature_count_table_dict:
        reads = ref_feature_count_table_dict[key][1:]
        concentration = new_dict_for_conc_table[key][mix]
        for gradient, val in reads.iteritems():
            if gradient not in combined_dict:
                combined_dict[gradient] = {}
            combined_dict[gradient][key] = val
        concentration_list.append(concentration)
    concentration_serie = pd.Series(concentration_list)
    combined_df = pd.DataFrame.from_dict(combined_dict)
    combined_df.loc[:, 'Concentration'] = concentration_serie.values
    combined_df.reset_index(inplace=True)
    return combined_df


def modify_input(gradient_file):
    """Take only numeric values"""
    read_grad_value = gradient_file[list(filter(
        lambda col: col.startswith("Grad"), gradient_file.columns))]
    read_concentration_value = gradient_file[list(filter(
        lambda col: col.startswith("Concentration"), gradient_file.columns))]
    return read_grad_value, read_concentration_value, gradient_file


def regression(read_grad_value, read_concentration_value, gradient_file,
               number_of_outliers):
    """Find the robust regression of the concentration and reads values"""
    table_with_outliers = {}
    key = gradient_file.iloc[:, 0].ravel()
    concentration_log10 = np.log10(read_concentration_value.values)
    for gradient in read_grad_value:
        read_grad_value_log10 = np.log10(read_grad_value[gradient])
        read_grad_value_log10[read_grad_value_log10 == -inf] = 0
        concentration_log10[concentration_log10 == -inf] = 0
        x_values = read_grad_value_log10.values.reshape(len(key), 1)
        y_values = concentration_log10.reshape(len(key), 1)
        clt_ransac = linear_model.RANSACRegressor(linear_model.LinearRegression())
        clt_ransac = clt_ransac.fit(x_values, y_values)
        fig, axes = plt.subplots(ncols=2, figsize=(15, 10))
        axes[0].scatter(x_values, y_values, c='g')
        axes[0].set_xlabel('Reads', fontsize=18)
        axes[0].set_ylabel('Concentration', fontsize=18)
        """Plot the robust regression, including inliers and outliers"""
        plt.plot(x_values, clt_ransac.predict(x_values), color='blue')
        robust_estimator = RANSACRegressor(random_state=0)
        robust_estimator.fit(x_values, y_values)
        distance_pred = robust_estimator.predict(x_values)
        mean_squared_error = (y_values - distance_pred) ** 2
        index = np.argsort(mean_squared_error.ravel())
        axes[1].scatter(x_values[index[:-number_of_outliers]],
                        y_values[index[:-number_of_outliers]], c='b', label='inliers',  alpha=0.5)
        axes[1].scatter(x_values[index[-number_of_outliers:]],
                        y_values[index[-number_of_outliers:]], c='r', label='outliers', alpha=0.5)
        axes[1].set_xlabel('Reads', fontsize=18)
        axes[1].legend(loc=2)
        plt.title(gradient + '\n'
                             "The intercept values is " +
                  str(clt_ransac.estimator_.intercept_)
                  + '\n' +
                  "The slope values is " +
                  str(clt_ransac.estimator_.coef_),
                  fontsize=15)
        plt.savefig(gradient)
        plt.close()
        """Find the outliers using the number that you defined as parameter"""
        """Selecting first the number of outliers that you want to find in the dataset"""
        outliers_keys = key[index[-number_of_outliers:]]
        table_with_outliers[gradient] = outliers_keys
    outliers_df = pd.DataFrame(table_with_outliers)
    return outliers_df


def common_ercc(csv_file):
    unique_vals = pd.Series(csv_file.values.ravel()).unique()
    data_dict = {}
    for val in unique_vals:
        row = []
        for value in csv_file.columns:
            row.append((csv_file[value] == val).sum())
        data_dict[val] = row
    """Selecting the cut-off"""
    results_df = pd.DataFrame(data_dict).T
    results_df.columns = csv_file.columns
    results_df = results_df.sum(axis=1).sort_values(ascending=False)
    return results_df


def new_ref_table(csv_common_ercc, ref_table, output_file, number_of_ercc_in_common):
    csv_common_ercc_df = pd.Series.to_frame(csv_common_ercc)
    new_csv_common_ercc_df = csv_common_ercc_df.reset_index()
    new_csv_common_ercc_df.columns = ['keys', 'values']
    selected_df = new_csv_common_ercc_df[~(new_csv_common_ercc_df['values'] < number_of_ercc_in_common)]
    selected_df.reset_index(drop=True, inplace=True)
    ref_table_df = pd.DataFrame.from_dict(ref_table, orient='index')
    my_keys = selected_df['keys'].tolist()
    for index, row in ref_table_df.iterrows():
        if index in my_keys:
            df = ref_table_df.drop(my_keys)
    df.to_csv(output_file, sep='\t', index=None)



import pandas as pd
import numpy as np


def scaling_(feature_count_table, feature_count_start_column, feature_count_end_column,
             pseudo_count, scaling_method, output_file):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column, feature_count_end_column)
    attribute_matrix = _extract_attributes(feature_count_table_df,
                                           feature_count_start_column)
    normalized_table = normalize_values(value_matrix, scaling_method, pseudo_count)
    pd.concat([attribute_matrix, normalized_table],
              axis=1).to_csv(output_file, sep='\t', index=0)


def _extract_value_matrix(feature_count_table_df, feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:, feature_count_start_column:(
        feature_count_end_column)]


def _extract_attributes(feature_count_table_df,
                        feature_count_start_column):
    return feature_count_table_df.iloc[:, : feature_count_start_column]


def normalize_values(values_matrix, scaling_method, pseudo_count):
    if scaling_method == "no_normalization":
        normalized_values = values_matrix
    elif scaling_method == "log2":
        normalized_values = values_matrix.applymap(
            lambda val: val + pseudo_count).applymap(np.log2)
    elif scaling_method == "log10":
        normalized_values = values_matrix.applymap(
            lambda val: val + pseudo_count).applymap(np.log10)
    elif scaling_method == "to_max":
        values_matrix = values_matrix.fillna(lambda x: 0)
        row_max_values = values_matrix.max(axis=1)
        normalized_values = values_matrix.divide(
            row_max_values, axis=0)
        normalized_values = pd.DataFrame(normalized_values).fillna(0)
    elif scaling_method == "to_range":
        row_max_values = values_matrix.max(axis=1)
        row_min_values = values_matrix.min(axis=1)
        normalized_values = values_matrix.subtract(
            row_min_values, axis=0).divide(row_max_values, axis=0)
    else:
        print("Normalization method not known")
    return normalized_values



import pandas as pd
import holoviews as hv
import matplotlib.pyplot as plt

hv.extension('bokeh', width=90)


def plot_heatmap(feature_count_table, feature_count_start_column,
                 feature_count_end_column, y_label_, output_file):
    table = pd.read_table(feature_count_table, sep="\t")
    table.set_index('Gene', inplace=True)
    value_matrix = _extract_value_matrix(table, feature_count_start_column,
                                         feature_count_end_column)
    heatmaps(value_matrix, y_label_, output_file)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column) - 1:feature_count_end_column - 1]


def heatmaps(value_matrix, y_label_, output_file):
    fig, axes_x = plt.subplots(figsize=(100, 100))
    imshow_ = axes_x.imshow(value_matrix, cmap="Greens")
    fig.colorbar(imshow_, ax=axes_x)
    axes_x.set_xticks(range(len(value_matrix.columns)))
    axes_x.set_yticks(range(len(value_matrix.index)))
    axes_x.set_xticklabels(range(1, 21))
    axes_x.set_yticklabels(value_matrix.index)
    axes_x.set_xlabel("Fractions", fontsize=18)
    axes_x.set_ylabel(y_label_, fontsize=10)
    plt.savefig(output_file)
    plt.show()



import pandas as pd
import umap
import numpy as np
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA

from bokeh.plotting import figure, ColumnDataSource, show, output_file
from bokeh.models import HoverTool, BoxZoomTool, ResetTool, PanTool, DataTable, TableColumn
from bokeh.models import WheelZoomTool, CustomJS, Button, TapTool, OpenURL
from bokeh.layouts import row, column


def interactive_plot(table_with_clusters, feature_count_start_column, feature_count_end_column,
                     dimension_reduction_algorithm, perplexity, n_neighbors, min_dist, url_link,
                     output_file_):
    read_counting_table = pd.read_csv(table_with_clusters, sep='\t')
    value_matrix = _extract_value_matrix(read_counting_table, feature_count_start_column,
                                         feature_count_end_column)
    if dimension_reduction_algorithm == "t-SNE":
        t_sne_result = perform_t_sne(value_matrix, perplexity)
        plot_t_sne(read_counting_table, t_sne_result, url_link, output_file_)
    elif dimension_reduction_algorithm == "PCA":
        pca_result = perform_pca(value_matrix)
        plot_pca(read_counting_table, pca_result, url_link, output_file_)
    elif dimension_reduction_algorithm == "UMAP":
        umap_result = perform_umap(value_matrix, n_neighbors, min_dist)
        plot_umap(read_counting_table, umap_result, url_link, output_file_)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):feature_count_end_column]


def perform_t_sne(normalized_values, perplexity):
    model = TSNE(n_components=2, random_state=0, perplexity=perplexity)
    np.set_printoptions(suppress=True)
    tsne_result = model.fit_transform(normalized_values)
    return tsne_result


def perform_pca(normalized_values):
    model = PCA(n_components=3, random_state=0)
    np.set_printoptions(suppress=True)
    pca_result = model.fit_transform(normalized_values)
    return pca_result


def perform_umap(normalized_values, n_neighbors, min_dist):
    reducer = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist)
    umap_results = reducer.fit_transform(normalized_values)
    return umap_results


def plot_t_sne(read_counting_table, tsne_result, url_link, output_file_):
    read_counting_table["t-SNE-component_1"] = [pos[0] for pos in tsne_result]
    read_counting_table["t-SNE-component_2"] = [pos[1] for pos in tsne_result]

    read_counting_table["Attributes_split"] = \
        read_counting_table["Attributes"].apply(lambda attr: dict(
            [key_value_pair.split("=")
             for key_value_pair in attr.split(";")]))

    hower_data = dict(
        x=read_counting_table["t-SNE-component_1"],
        y=read_counting_table["t-SNE-component_2"],
        feature=read_counting_table["Feature"],
        cluster_label=read_counting_table["Cluster_label"],
        gene=read_counting_table["Gene"],
        attribute=read_counting_table["Attributes"])

    source = ColumnDataSource(hower_data)
    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("Feature", "@feature"),
        ("Cluster label", "@cluster_label")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "lasso_select", "tap"],
                  output_backend="webgl", lod_threshold=100,
                  title="Grad-Seq t-SNE RNA-Seq")
    plot.toolbar.logo = None
    plot.circle("x", "y", source=source, size=7, alpha=3, line_color="grey", color="grey")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'
    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"
    source2 = ColumnDataSource(data=dict(x=[], y=[]))
    plot2 = figure(plot_width=300, plot_height=300,
                   tools="", title="Zoom", output_backend="webgl", lod_threshold=100)
    plot2.toolbar.logo = None
    plot2.circle('x', 'y', source=source2, alpha=0.6)
    columns = [
        TableColumn(field="gene", title="Genes"),
        TableColumn(field="feature", title="Feature"),
        TableColumn(field="attribute", title="Attributes")
    ]
    data_table = DataTable(source=source, columns=columns, width=450, height=800, fit_columns=True)
    savebutton = Button(label="Save", button_type="success")
    source.selected.js_on_change('indices', CustomJS(args=dict(s1=source, s2=source2), code="""
            var inds = cb_obj.indices;
            var d1 = s1.data;
            var d2 = s2.data;
            d2['x'] = []
            d2['y'] = []
            for (var i = 0; i < inds.length; i++) {
                d2['x'].push(d1['x'][inds[i]])
                d2['y'].push(d1['y'][inds[i]])
            }
            s2.change.emit();
        """)
                                 )

    savebutton.callback = CustomJS(args=dict(source_data=source), code="""
        var inds = source_data.selected['1d'].indices;
        var data = source_data.data;
        console.log("inds", inds);
        console.log("data", data);
        var out = "Features\\tGene\\tAttributes + \\n";
        for (i = 0; i < inds.length; i++) {
                 out += data['feature'][inds[i]] + "\\t" + data['gene'][inds[i]] + 
                 "\\t" +  data['attribute'][inds[i]] + "\\n";
        }
        var file = new Blob([out], {type: 'text/plain'});
        var elem = window.document.createElement('a');
        elem.href = window.URL.createObjectURL(file);
        elem.download = 'selected-data.csv';
        document.body.appendChild(elem);
        elem.click();
        document.body.removeChild(elem);
        """)

    output_file(output_file_ + ".html", title="Grad-seq t-SNE")
    layout = row(plot, data_table, column(plot2, savebutton))
    show(layout)


def plot_pca(read_counting_table, pca_result, url_link, output_file_):
    read_counting_table["PCA-component_1"] = [pos[0] for pos in pca_result]
    read_counting_table["PCA-component_2"] = [pos[1] for pos in pca_result]

    read_counting_table["Attributes_split"] = \
        read_counting_table["Attributes"].apply(lambda attr: dict(
            [key_value_pair.split("=")
             for key_value_pair in attr.split(";")]))

    hower_data = dict(
        x=read_counting_table["PCA-component_1"],
        y=read_counting_table["PCA-component_2"],
        feature=read_counting_table["Feature"],
        cluster_label=read_counting_table["Cluster_label"],
        gene=read_counting_table["Gene"],
        attribute=read_counting_table["Attributes"])

    source = ColumnDataSource(hower_data)
    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("Feature", "@feature"),
        ("Cluster label", "@cluster_label")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "lasso_select", "tap"],
                  output_backend="webgl", lod_threshold=100,
                  title="Grad-Seq PCA RNA-Seq")
    plot.toolbar.logo = None
    plot.circle("x", "y", source=source, size=7, alpha=3, line_color="grey", color="grey")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'
    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Component 1"
    plot.yaxis.axis_label = "Component 2"
    source2 = ColumnDataSource(data=dict(x=[], y=[]))
    plot2 = figure(plot_width=300, plot_height=300,
                   tools="", title="Zoom", output_backend="webgl", lod_threshold=100)
    plot2.toolbar.logo = None
    plot2.circle('x', 'y', source=source2, alpha=0.6)
    columns = [
        TableColumn(field="gene", title="Genes"),
        TableColumn(field="feature", title="Feature"),
        TableColumn(field="attribute", title="Attributes")
    ]
    data_table = DataTable(source=source, columns=columns, width=450, height=800, fit_columns=True)
    savebutton = Button(label="Save", button_type="success")
    source.selected.js_on_change('indices', CustomJS(args=dict(s1=source, s2=source2), code="""
            var inds = cb_obj.indices;
            var d1 = s1.data;
            var d2 = s2.data;
            d2['x'] = []
            d2['y'] = []
            for (var i = 0; i < inds.length; i++) {
                d2['x'].push(d1['x'][inds[i]])
                d2['y'].push(d1['y'][inds[i]])
            }
            s2.change.emit();
        """)
                                 )

    savebutton.callback = CustomJS(args=dict(source_data=source), code="""
        var inds = source_data.selected['1d'].indices;
        var data = source_data.data;
        console.log("inds", inds);
        console.log("data", data);
        var out = "Features\\tGene\\tAttributes + \\n";
        for (i = 0; i < inds.length; i++) {
                 out += data['feature'][inds[i]] + "\\t" + data['gene'][inds[i]] + "\\t" +  
                 data['attribute'][inds[i]] + "\\n";
        }
        var file = new Blob([out], {type: 'text/plain'});
        var elem = window.document.createElement('a');
        elem.href = window.URL.createObjectURL(file);
        elem.download = 'selected-data.csv';
        document.body.appendChild(elem);
        elem.click();
        document.body.removeChild(elem);
        """)

    output_file(output_file_ + ".html", title="Grad-seq PCA")
    layout = row(plot, data_table, column(plot2, savebutton))
    show(layout)


def plot_umap(read_counting_table, umap_result, url_link, output_file_):
    read_counting_table["UMAP-component_1"] = [pos[0] for pos in umap_result]
    read_counting_table["UMAP-component_2"] = [pos[1] for pos in umap_result]

    read_counting_table["Attributes_split"] = \
        read_counting_table["Attributes"].apply(lambda attr: dict(
            [key_value_pair.split("=")
             for key_value_pair in attr.split(";")]))

    hower_data = dict(
        x=read_counting_table["UMAP-component_1"],
        y=read_counting_table["UMAP-component_2"],
        feature=read_counting_table["Feature"],
        cluster_label=read_counting_table["Cluster_label"],
        gene=read_counting_table["Gene"],
        attribute=read_counting_table["Attributes"])

    source = ColumnDataSource(hower_data)
    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("Feature", "@feature"),
        ("Cluster label", "@cluster_label")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "lasso_select", "tap"],
                  output_backend="webgl", lod_threshold=100,
                  title="Grad-Seq UMAP RNA-Seq")
    plot.toolbar.logo = None
    plot.circle("x", "y", source=source, size=7, alpha=3, line_color="grey", color="grey")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'
    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"
    source2 = ColumnDataSource(data=dict(x=[], y=[]))
    plot2 = figure(plot_width=300, plot_height=300,
                   tools="", title="Zoom", output_backend="webgl", lod_threshold=100)
    plot2.toolbar.logo = None
    plot2.circle('x', 'y', source=source2, alpha=0.6)
    columns = [
        TableColumn(field="gene", title="Genes"),
        TableColumn(field="feature", title="Feature"),
        TableColumn(field="attribute", title="Attributes")
    ]
    data_table = DataTable(source=source, columns=columns, width=450, height=800, fit_columns=True)
    savebutton = Button(label="Save", button_type="success")
    source.selected.js_on_change('indices', CustomJS(args=dict(s1=source, s2=source2), code="""
            var inds = cb_obj.indices;
            var d1 = s1.data;
            var d2 = s2.data;
            d2['x'] = []
            d2['y'] = []
            for (var i = 0; i < inds.length; i++) {
                d2['x'].push(d1['x'][inds[i]])
                d2['y'].push(d1['y'][inds[i]])
            }
            s2.change.emit();
        """)
                                 )

    savebutton.callback = CustomJS(args=dict(source_data=source), code="""
        var inds = source_data.selected['1d'].indices;
        var data = source_data.data;
        console.log("inds", inds);
        console.log("data", data);
        var out = "Features\\tGene\\tAttributes + \\n";
        for (i = 0; i < inds.length; i++) {
                 out += data['feature'][inds[i]] + "\\t" + data['gene'][inds[i]] + "\\t" +  
                 data['attribute'][inds[i]] + "\\n";
        }
        var file = new Blob([out], {type: 'text/plain'});
        var elem = window.document.createElement('a');
        elem.href = window.URL.createObjectURL(file);
        elem.download = 'selected-data.csv';
        document.body.appendChild(elem);
        elem.click();
        document.body.removeChild(elem);
        """)

    output_file(output_file_ + ".html", title="Grad-seq UMAP")
    layout = row(plot, data_table, column(plot2, savebutton))
    show(layout)



import pandas as pd


def dropping_column(feature_count_table, column_to_drop, output_file):
    feature_count_table_df = pd.read_table(feature_count_table)
    table_filtered_gene_quanti = create_gene_quanti_table_modified(feature_count_table_df,
                                                                   column_to_drop)
    table_filtered_gene_quanti.to_csv(output_file, sep='\t', index=None)


def create_gene_quanti_table_modified(gene_quanti_table, column_to_drop):
    cols = gene_quanti_table.columns.tolist()
    gene_quanti_table_with_gene_column_modified = gene_quanti_table[cols]
    gene_quanti_table_with_gene_column_modified.drop(column_to_drop, axis=1,
                                                     inplace=True)
    return gene_quanti_table_with_gene_column_modified



import pandas as pd


def normalized_count_table_spike_in(feature_count_table, feature_count_start_column, feature_count_end_column,
                                    size_factors_table, name_column_with_values, normalized_table):
    feature_count_table_df = pd.read_csv(feature_count_table, sep='\t')
    size_factor_table_df = pd.read_csv(size_factors_table, sep='\t')
    normalized_table_df = _normalize_by_size_factor(feature_count_table_df, feature_count_start_column,
                                                    feature_count_end_column,
                                                    size_factor_table_df, name_column_with_values)
    normalized_table_df.to_csv(normalized_table, sep='\t', index=None)


def _normalize_by_size_factor(feature_counting_table_df,
                              feature_count_start_column, feature_count_end_column, size_factor_table,
                              name_column_with_values):
    df = feature_counting_table_df.copy()
    value_columns = df.iloc[:, int(feature_count_start_column
                                   ):int(feature_count_end_column)].columns
    series_column = pd.Series(size_factor_table[name_column_with_values].values)
    """Divide the gene quantification table for the size factor to normalize the data"""
    df[value_columns] = df[value_columns].divide(list(series_column), axis=1)
    return df



import os
from jinja2 import Template
import pandas as pd
from bokeh.plotting import figure, output_file, save
from bokeh.models import BoxZoomTool, ResetTool, PanTool
from bokeh.models import WheelZoomTool


def generate_html_page(html_folder, html_folder_gene, html_folder_protein, rna_seq_file, name_column_rna,
                       column_start_rna,
                       column_end_rna,
                       protein_file, name_column_protein,
                       column_start_protein, column_end_protein):
    image_folder_gene = "{}/images".format(html_folder_gene)
    image_folder_protein = "{}/images".format(html_folder_protein)
    rna_seq_data = pd.read_csv(rna_seq_file, sep="\t")
    protein_data = pd.read_csv(protein_file, sep="\t")
    for folder in [html_folder, html_folder_gene, html_folder_protein, image_folder_gene, image_folder_protein]:
        if not os.path.exists(folder):
            os.mkdir(folder)
    generate_index_pager(rna_seq_data, protein_data, html_folder)
    generate_gene_pages(rna_seq_data, name_column_rna, column_start_rna, column_end_rna, html_folder_gene,
                        image_folder_gene)
    generate_protein_pages(protein_data, name_column_protein, column_start_protein, column_end_protein,
                           html_folder_protein, image_folder_protein)


def generate_index_pager(rna_seq_data, protein_data, html_folder):
    # template = Template(open("templates/gene_index_template.html").read())
    template = Template(open(
        "../templates/gene_index_template_bootstrap.html").read())
    gene_index_html = template.render(rna_seq_data=rna_seq_data, protein_data=protein_data)
    with open("{}/index.html".format(html_folder), "w") as output_fh:
        output_fh.write(gene_index_html)


def generate_gene_pages(rna_seq_data, name_column_rna, column_start_rna, column_end_rna, html_folder_gene,
                        image_folder):
    for index, row in rna_seq_data.iterrows():
        generate_gene_page(row, name_column_rna, html_folder_gene, image_folder)
        generate_gene_rna_distri_image(row, name_column_rna, column_start_rna, column_end_rna, image_folder)
        # exit(0)


def generate_protein_pages(protein_data, name_column_protein, column_start_protein, column_end_protein,
                           html_folder_protein, image_folder):
    for index, row in protein_data.iterrows():
        generate_protein_page(row, name_column_protein, html_folder_protein, image_folder)
        generate_protein_distri_image(row, name_column_protein, column_start_protein, column_end_protein,
                                      image_folder)
        # exit(0)


def generate_gene_page(row, name_column_rna, html_folder, image_folder):
    locus_tag = row[name_column_rna]
    with open("{}/{}.html".format(html_folder, locus_tag),
              "w") as output_fh:
        # template = Template(open("templates/gene_template.html").read())
        template = Template(open("../templates/gene_template_bootstrap.html").read())
        gene_html = template.render(row=row)
        output_fh.write(gene_html.format(image_folder, locus_tag))


def generate_protein_page(row, name_column_protein, html_folder, image_folder):
    locus_tag = row[name_column_protein]
    with open("{}/{}.html".format(html_folder, locus_tag),
              "w") as output_fh:
        # template = Template(open("templates/gene_template.html").read())
        template = Template(open("../templates/protein_template_bootstrap.html").read())
        gene_html = template.render(row=row)
        output_fh.write(gene_html.format(image_folder, locus_tag))


def generate_gene_rna_distri_image(row, name_column_rna, column_start_rna, column_end_rna, image_folder):
    locus_tag = row[name_column_rna]
    counting_value_list = row.iloc[column_start_rna:column_end_rna]
    plot_html = figure(title=locus_tag, plot_width=600, plot_height=600,
                       x_axis_label='Fraction number',
                       y_axis_label='Normalized and scaled to max read counts',
                       tools=[BoxZoomTool(), ResetTool(), PanTool(),
                              WheelZoomTool()])
    plot_html.yaxis.axis_label_text_font_size = "15pt"
    plot_html.xaxis.axis_label_text_font_size = "15pt"
    plot_html.title.text_font_size = '15pt'
    plot_html.toolbar.logo = None
    y_axis = range(1, 21)
    plot_html.line(y_axis, counting_value_list)
    output_file("{}/{}.html".format(image_folder, locus_tag))
    save(plot_html)


def generate_protein_distri_image(row, name_column_protein, column_start_protein, column_end_protein, image_folder):
    locus_tag = row[name_column_protein]
    counting_value_list = row.iloc[column_start_protein:column_end_protein]
    plot_html = figure(title=locus_tag, plot_width=600, plot_height=600,
                       x_axis_label='Fraction number',
                       y_axis_label='Normalized and scaled to max read counts',
                       tools=[BoxZoomTool(), ResetTool(), PanTool(),
                              WheelZoomTool()])
    plot_html.yaxis.axis_label_text_font_size = "15pt"
    plot_html.xaxis.axis_label_text_font_size = "15pt"
    plot_html.title.text_font_size = '15pt'
    plot_html.toolbar.logo = None
    y_axis = range(1, 21)
    plot_html.line(y_axis, counting_value_list)
    output_file("{}/{}.html".format(image_folder, locus_tag))
    save(plot_html)



import pandas as pd
import scipy.stats.mstats


def normalized_count_table(
        feature_count_table,
        feature_count_start_column,
        feature_count_end_column,
        ref_feature_count_table,
        ref_feature_count_start_column,
        ref_feature_count_end_column,
        normalized_table,
        size_factor_table):
    ref_feature_count_table_df = pd.read_table(ref_feature_count_table, sep='\t')
    feature_count_table_df = pd.read_table(feature_count_table)
    size_factors = _calc_size_factors(ref_feature_count_table_df,
                                      ref_feature_count_start_column, ref_feature_count_end_column)
    normalized_table_df = _normalize_by_size_factor(feature_count_table_df,
                                                    feature_count_start_column,
                                                    feature_count_end_column, size_factors)
    normalized_table_df.to_csv(normalized_table, sep='\t', index=None)
    size_factors.to_csv(size_factor_table, sep='\t', header=['size_factors'])


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):feature_count_end_column]


def _calc_size_factors(ref_feature_count_table_df,
                       ref_feature_count_start_column, ref_feature_count_end_column):
    # TODO: Documentation the calculation with formula
    counting_value_matrix = _extract_value_matrix(
        ref_feature_count_table_df, ref_feature_count_start_column,
        int(ref_feature_count_end_column))
    geometric_means = _geometric_means(counting_value_matrix)
    scaled_counting_value_matrix = _multiply_geometric_means_with_value_matrix(
        counting_value_matrix, geometric_means)
    return scaled_counting_value_matrix.median(axis=0)


def _multiply_geometric_means_with_value_matrix(counting_value_matrix,
                                                geometric_means):
    """Before the multiplication remove rows where the geometric mean is 0"""
    nonzero = geometric_means != 0.0
    nonzero_geometric_mean = geometric_means[nonzero]
    nonzero_counting_value_matrix = counting_value_matrix[nonzero]
    return nonzero_counting_value_matrix.divide(nonzero_geometric_mean, axis=0)


def _geometric_means(counting_value_matrix):
    g_mean = scipy.stats.mstats.gmean(counting_value_matrix, axis=1)
    return g_mean


def _normalize_by_size_factor(feature_counting_table_df,
                              feature_count_start_column,
                              feature_count_end_column, size_factor):
    df = feature_counting_table_df.copy()
    value_columns = df.iloc[:, int(feature_count_start_column
                                   ):int(feature_count_end_column)].columns
    """Divide the gene quantification table for the size factor to normalize the data"""
    df[value_columns] = df[value_columns].divide(list(size_factor), axis=1)
    return df



import pandas as pd
import numpy as np


def merge_exctracted_attributes(feature_count_table, output_table):
    table = pd.read_csv(feature_count_table, sep='\t')
    table['Parent'] = np.where(table['Parent'].isnull(), table['name'], table['Parent'])
    table['Parent'] = np.where(table['Parent'].isnull(), table['ID'], table['Parent'])
    table.rename(columns={'Parent': 'Gene'}, inplace=True)
    table.to_csv(output_table, index=None, sep='\t')




import pandas as pd
from sklearn.manifold import TSNE
import numpy as np
import bokeh.palettes
from bokeh.plotting import figure, output_file, save, ColumnDataSource
from bokeh.models import HoverTool, BoxZoomTool, ResetTool, PanTool
from bokeh.models import WheelZoomTool, TapTool, OpenURL


def t_sne(feature_count_table, feature_count_start_column,
          feature_count_end_column,
          perplexity, srna_list_files, cluster_names, url_link,
          output_file_colorized_by_clusters,
          output_file_colorized_by_rna_class,
          output_colored_by_lists):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column,
                                         feature_count_end_column)
    t_sne_results = perform_t_sne(value_matrix, perplexity)
    plot_t_sne_using_clustering(feature_count_table_df, t_sne_results,
                                output_file_colorized_by_clusters, url_link)
    plot_using_only_rna_colors(feature_count_table_df, t_sne_results,
                               output_file_colorized_by_rna_class, url_link)

    if srna_list_files:
        srnas_and_list_names = read_srna_lists(srna_list_files)
        plot_t_sne_colored_by_lists(feature_count_table_df, t_sne_results, output_colored_by_lists,
                                    srnas_and_list_names, cluster_names, url_link)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):feature_count_end_column]


def perform_t_sne(normalized_values, perplexity):
    model = TSNE(n_components=2, random_state=0, perplexity=perplexity)
    np.set_printoptions(suppress=True)
    tsne_result = model.fit_transform(normalized_values)
    return tsne_result


def plot_t_sne_using_clustering(read_counting_table, tsne_result,
                                output_file_colorized_by_clusters, url_link):
    read_counting_table["t-SNE-component_1"] = [pos[0] for pos in tsne_result]
    read_counting_table["t-SNE-component_2"] = [pos[1] for pos in tsne_result]

    read_counting_table["Attributes_split"] = \
        read_counting_table["Attributes"].apply(lambda attr: dict(
            [key_value_pair.split("=")
             for key_value_pair in attr.split(";")]))

    color_palette = bokeh.palettes.Colorblind[(
        len(read_counting_table["Cluster_label"].unique()))]
    color = read_counting_table["Cluster_label"].apply(
        lambda lable: color_palette[lable])
    label = read_counting_table.apply(
        _label_clustering, axis=1)

    hower_data = dict(
        x=read_counting_table["t-SNE-component_1"],
        y=read_counting_table["t-SNE-component_2"],
        feature=read_counting_table["Feature"],
        gene=read_counting_table['Gene'],
        cluster_label=read_counting_table["Cluster_label"],
        color=color,
        label=label)

    hower_data_df = pd.DataFrame.from_dict(hower_data)
    hower_data_df.to_csv("hower_data_df.csv", sep='\t', index=None)

    for feature in ["ID", "Parent", "Dbxref", "Note", "function",
                    "gbkey", "product", "sRNA_type"]:
        read_counting_table[feature] = read_counting_table[
            "Attributes_split"].apply(
            lambda attributes: attributes.get(feature, "-"))
        hower_data[feature] = read_counting_table[feature]

    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("ID", "@ID"),
        ("Feature", "@feature"),
        ("Cluster label", "@cluster_label")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq t-SNE RNA-Seq")

    plot.circle("x", "y", source=source, size=7, alpha=3, color='color',
                legend="label", line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'
    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"

    output_file(output_file_colorized_by_clusters)
    save(plot)


def _label_clustering(row):
    __label = {0: "cluster1", 1: "cluster2", 2: "cluster3",
               3: "cluster4", 4: "cluster5", 5: "cluster6",
               6: "cluster7", 7: "cluster8"}[row["Cluster_label"]]
    return __label


def create_palette_map(read_counting_table):
    feature_unique_values = read_counting_table["Feature"].unique()
    colors = len(feature_unique_values)

    if colors < 3:
        color_palette = ["#0000ff", "#ffff00"]
    else:
        color_palette = bokeh.palettes.Colorblind[colors]

    palette_map = {}
    for index in range(0, len(feature_unique_values)):
        palette_map[feature_unique_values[index]] = color_palette[index]
    return palette_map


def plot_using_only_rna_colors(read_counting_table, t_sne_result,
                               output_file_colorized_by_rna_class, url_link):
    read_counting_table["t-SNE-component_1"] = [pos[0] for pos in t_sne_result]
    read_counting_table["t-SNE-component_2"] = [pos[1] for pos in t_sne_result]

    read_counting_table["Attributes_split"] = read_counting_table[
        "Attributes"].apply(
        lambda attr: dict(
            [key_value_pair.split("=") for
             key_value_pair in attr.split(";")]))

    palette_map = create_palette_map(read_counting_table)
    color = read_counting_table["Feature"].apply(
        lambda lable: palette_map[lable])
    label = read_counting_table.apply(
        _label, axis=1)

    hower_data = dict(
        x=read_counting_table["t-SNE-component_1"],
        y=read_counting_table["t-SNE-component_2"],
        feature=read_counting_table["Feature"],
        gene=read_counting_table["Gene"],
        color=color,
        label=label)

    for feature in ["product", "ID", "type", "ncrna_class",
                    "sRNA_type", "Name", "pseudo"]:
        read_counting_table[feature] = read_counting_table[
            "Attributes_split"].apply(
            lambda attributes: attributes.get(feature, "-"))
        hower_data[feature] = read_counting_table[feature]

    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("ID", "@ID"),
        ("Feature", "@feature")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq t-SNE RNA-Seq")

    plot.circle("x", "y", source=source, size=7, alpha=3, color='color',
                legend='label', line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'

    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"

    output_file(output_file_colorized_by_rna_class)
    save(plot)


def _label(row):
    label = {"CDS": "CDS", "ncRNA": "ncRNA", "tRNA": "tRNA", "gene": "gene",
             "rRNA": "rRNA", "tmRNA": "tmRNA", "5UTR": "5UTR",
             "3UTR": "3UTR", "sRNA": "sRNA"
             }[row["Feature"]]
    return label


def read_srna_lists(srna_list_files):
    srnas_and_list_names = {}
    for index, srna_list_file in enumerate(srna_list_files):
        list_name = "sRNA_cluster_{}".format(index + 1)
        for line in open(srna_list_file):
            srnas_and_list_names[line.strip()] = list_name
    return srnas_and_list_names


def plot_t_sne_colored_by_lists(read_counting_table, tsne_result,
                                output_file_list, srnas_and_list_names,
                                cluster_names, url_link):
    read_counting_table["t-SNE-component_1"] = [pos[0] for pos in tsne_result]
    read_counting_table["t-SNE-component_2"] = [pos[1] for pos in tsne_result]

    read_counting_table["Attributes_split"] = read_counting_table[
        "Attributes"].apply(
        lambda attr: dict(
            [key_value_pair.split("=") for
             key_value_pair in attr.split(";")]))

    palette_map = create_palette_map(read_counting_table)
    color = read_counting_table.apply(
        _color_1, args=(srnas_and_list_names, palette_map), axis=1)
    label = read_counting_table.apply(
        _label_1, args=(srnas_and_list_names, cluster_names), axis=1)

    hower_data = dict(
        x=read_counting_table["t-SNE-component_1"],
        y=read_counting_table["t-SNE-component_2"],
        feature=read_counting_table["Feature"],
        gene=read_counting_table["Gene"],
        color=color,
        label=label)

    for feature in ["product", "ID", "type", "ncrna_class",
                    "sRNA_type", "Name", "pseudo"]:
        read_counting_table[feature] = read_counting_table[
            "Attributes_split"].apply(
            lambda attributes: attributes.get(feature, "-"))
        hower_data[feature] = read_counting_table[feature]

    source = ColumnDataSource(hower_data)

    hover = HoverTool(tooltips=[
        ("Gene", "@gene"),
        ("ID", "@ID"),
        ("Feature", "@feature")])

    plot = figure(plot_width=900, plot_height=900,
                  tools=[hover, BoxZoomTool(), ResetTool(), PanTool(),
                         WheelZoomTool(), "tap"],
                  title="Grad-Seq t-SNE RNA-Seq")

    plot.circle("x", "y", source=source, size=7, alpha=3, color="color",
                legend='label', line_color="black")
    plot.yaxis.axis_label_text_font_size = "15pt"
    plot.xaxis.axis_label_text_font_size = "15pt"
    plot.title.text_font_size = '15pt'

    url = url_link
    taptool = plot.select(type=TapTool)
    taptool.callback = OpenURL(url=url)

    plot.xaxis.axis_label = "Dimension 1"
    plot.yaxis.axis_label = "Dimension 2"

    output_file(output_file_list)
    save(plot)


def _color_1(row, srnas_and_list_names, palette_map):
    color = palette_map[row["Feature"]]
    srna_cluster_color = {"sRNA_cluster_1": "#FF4D4D",
                          "sRNA_cluster_2": "#F0F3F4",
                          "sRNA_cluster_3": "#000000",
                          "sRNA_cluster_4": "#FFFF00"}
    for feature in ["Gene"]:
        if row[feature] in srnas_and_list_names:
            color = srna_cluster_color[
                srnas_and_list_names[row[feature]]]
    return color


def _label_1(row, srnas_and_list_names, cluster_names):
    label = {"ncRNA": "other ncRNAs", "sRNA": "other ncRNAs",
             "CDS": "CDS", "tRNA": "tRNA",
             "rRNA": "rRNA", "5UTR": "5UTR", "3UTR":
                 "3UTR", "gene": "gene"}[row["Feature"]]
    srna_cluster_label = {}
    for index in range(0, len(cluster_names)):
        srna_cluster_label["sRNA_cluster_" + str(index + 1)] = \
            cluster_names[index]

    for feature in ["Gene"]:
        if row[feature] in srnas_and_list_names:
            label = srna_cluster_label[
                srnas_and_list_names[row[feature]]]
    return label



import matplotlib.pyplot as plt
import pandas as pd


def plot_histograms(feature_count_table,
                    feature_count_start_column):
    feature_count_table_df = pd.read_table(feature_count_table)
    feature_count_table_df_value = _extract_value_matrix(feature_count_table_df,
                                                         feature_count_start_column)
    _plot_histogram_pdf(feature_count_table_df_value)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):]


def _plot_histogram_pdf(feature_count_table_df_value):
    i = 0
    for column in feature_count_table_df_value:
        i = i + 1
        plt.figure(i)
        plt.hist(feature_count_table_df_value[column].values)
        # plt.show()
        plt.savefig(column)



import pandas as pd


def size_factor_extraction(ref_feature_count_table, size_factor_table):
    table = pd.read_table(ref_feature_count_table, sep='\t')
    filtered_table = create_a_new_table_filtered_ercc(table).transpose()
    filtered_table.to_csv(size_factor_table, sep='\t')


def create_a_new_table_filtered_ercc(alignment_table):
    series = []
    for index, row in alignment_table.iterrows():
        selection = row['Libraries']
        if selection.startswith('spike_1') & ('No. of aligned reads' in selection):
            series.append(row)

    filtered_table = pd.DataFrame(series)
    values_matrix = filtered_table.fillna(lambda x: 0)
    row_max_values = int(values_matrix.max(axis=1))
    values_matrix_index = values_matrix.set_index('Libraries')
    normalized_values = values_matrix_index.divide(
        row_max_values, axis=0)
    return normalized_values



import pandas as pd
from scipy.stats import spearmanr
import matplotlib.pyplot as plt


def correlation_replicates_sd(table_replicate1, table_replicate2, table_start_column,
                              table_end_column, output_table,
                              output_figure):
    table_1r = pd.read_csv(table_replicate1, sep="\t")
    table_2r = pd.read_csv(table_replicate2, sep="\t")

    table_1r.set_index('Gene', inplace=True)
    table_1r.index = table_1r.index + '_' + table_1r.groupby(level=0).cumcount().astype(str)

    table_2r.set_index('Gene', inplace=True)
    table_2r.index = table_2r.index + '_' + table_2r.groupby(level=0).cumcount().astype(str)
    no_in_1 = table_2r[~table_2r.index.isin(table_1r.index)]
    index_no_in_1 = no_in_1.index
    table_2r.drop(index_no_in_1, inplace=True)

    value_matrix_table_1r = _extract_value_matrix(table_1r, table_start_column,
                                                  table_end_column)
    attribute_matrix_1r = attribute_matrix(table_1r, table_start_column)
    value_matrix_table_2r = _extract_value_matrix(table_2r, table_start_column,
                                                  table_end_column)
    concat_df = correlation(value_matrix_table_1r, value_matrix_table_2r)
    plot(concat_df, output_figure)
    pd.concat([attribute_matrix_1r, concat_df],
              axis=1).to_csv(output_table, sep='\t')


def _extract_value_matrix(feature_count_table_df, feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):int(
        feature_count_end_column)]


def attribute_matrix(feature_count_table_df, feature_count_start_column):
    return feature_count_table_df.iloc[:, : int(feature_count_start_column)]


def correlation(value_matrix_1r, value_matrix_2r):
    concat_dataframe = pd.DataFrame()
    for index_r1, row_r1 in value_matrix_1r.iterrows():
        row_r2 = value_matrix_2r.loc[index_r1]
        concat_row = pd.concat([row_r1, row_r2])
        rho = spearmanr(row_r1, row_r2, axis=None)[0]
        p_value = spearmanr(row_r1, row_r2, axis=None)[1]
        concat_row.at["rho"] = rho
        concat_row.at["pvalue"] = p_value
        concat_dataframe = concat_dataframe.append(concat_row)[concat_row.keys().tolist()]

    return concat_dataframe


def plot(concat_df, output_figure):
    plt.style.use('ggplot')
    fig = plt.figure()
    axes_x = fig.add_subplot(1, 1, 1)
    axes_x.set_xlim(concat_df["rho"].min(), 1)
    plt.hist(concat_df["rho"], bins=50)
    plt.title("Spearman correlation coefficient between replicates (RNA-Seq)",
              fontsize=13)
    plt.xlabel("Spearman correlation coefficient", fontsize=15)
    plt.ylabel("Occurance", fontsize=15)
    plt.savefig(output_figure)



import pandas as pd
import matplotlib.pyplot as plt
from bokeh.plotting import figure, output_file, show
from bokeh.models import BoxZoomTool, ResetTool, PanTool
from bokeh.models import WheelZoomTool


def plot_kinetics(feature_count_table, feature_count_start_column, feature_count_end_column,
                  gene_name, output_format):
    feature_count_table_df = pd.read_table(feature_count_table)

    gene_row = _extract_gene_row(feature_count_table_df, gene_name)
    value_gene_row = _extract_value_gene_row(gene_row, feature_count_start_column,
                                             feature_count_end_column).values.T.tolist()
    if output_format == "pdf":
        _plot_gene_png(value_gene_row, gene_name)
    else:
        _plot_gene_html(value_gene_row, gene_name)


def _extract_gene_row(feature_count_table_df, gene_name):
    return feature_count_table_df.loc[feature_count_table_df['Gene'] == gene_name]


def _extract_value_gene_row(gene_row, feature_count_start_column, feature_count_end_column):
    return gene_row.iloc[:, feature_count_start_column:feature_count_end_column]


def _plot_gene_html(counting_value_list, gene_name):
    output_file(gene_name + '.html')
    plot_html = figure(title=gene_name,
                       x_axis_label='Fraction number',
                       y_axis_label='Normalized and scaled to max read counts',
                       tools=[BoxZoomTool(), ResetTool(), PanTool(),
                              WheelZoomTool()])
    plot_html.yaxis.axis_label_text_font_size = "15pt"
    plot_html.xaxis.axis_label_text_font_size = "15pt"
    plot_html.title.text_font_size = '15pt'
    plot_html.toolbar.logo = None
    y_axis = range(1, 22)
    plot_html.line(y_axis, counting_value_list)
    show(plot_html)


def _plot_gene_png(counting_value_list, gene_name):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(counting_value_list)
    plt.grid(True)
    plt.xlabel('Fraction number', fontsize=15)
    plt.ylabel('Normalized and scaled to max read counts', fontsize=15)
    plt.title(gene_name, fontsize=18)
    plt.savefig(gene_name, format='pdf')



import pandas as pd
from sklearn.neighbors import NearestNeighbors

import matplotlib.pyplot as plt


def generate_nearest_neighbors(feature_count_table,
                               feature_count_start_column):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column)
    nearest_neighbors(value_matrix)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):]


def _extract_attributes(feature_count_table_df,
                        feature_count_start_column):
    return feature_count_table_df.iloc[:, : int(feature_count_start_column)]


def nearest_neighbors(value_matrix):
    nearest_neighbors_ = NearestNeighbors(n_neighbors=41).fit(value_matrix)
    distances, indices = nearest_neighbors_.kneighbors(value_matrix)

    my_distance = distances[:, -1]
    my_distance_df = pd.DataFrame(my_distance)
    my_distance_df_div = my_distance_df.divide(len(my_distance))
    sorted_df = my_distance_df_div.sort_values(by=0, ascending=False)
    sorted_list = sorted_df[0].tolist()
    plt.plot(sorted_list)
    plt.show()



import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist, pdist
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt


def k_means_clustering_elbow(feature_count_table, feature_count_start_column,
                             feature_count_end_column,
                             min_number_of_clusters, max_number_of_clusters, output_plots1,
                             output_plot2):
    feature_count_table_df = pd.read_csv(feature_count_table, sep='\t')
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column,
                                         feature_count_end_column)
    elbow_methods(value_matrix, min_number_of_clusters, max_number_of_clusters,
                  output_plots1)
    elbow_methods1(value_matrix, min_number_of_clusters,
                   max_number_of_clusters, output_plot2)


def elbow_methods(value_matrix, min_number_of_clusters,
                  max_number_of_clusters, output_plots1):
    range_k = range(int(min_number_of_clusters), int(max_number_of_clusters))
    k_m = [KMeans(n_clusters=k).fit(value_matrix) for k in range_k]
    centroids = [k.cluster_centers_ for k in k_m]
    d_k = [cdist(value_matrix, cent, 'euclidean') for cent in centroids]
    dist = [np.min(D, axis=1) for D in d_k]
    w_css = [sum(d ** 2) for d in dist]
    tss = sum(pdist(value_matrix) ** 2) / value_matrix.shape[0]

    bss = tss - w_css
    fig = plt.figure()
    axes_x = fig.add_subplot(111)
    axes_x.plot(range_k, bss / tss * 100, 'b*-')
    plt.grid(True)
    plt.xlabel('Number of clusters', fontsize=15)
    plt.ylabel('Percentage of variance explained', fontsize=15)
    plt.title('Elbow for KMeans clustering', fontsize=18)
    plt.savefig(output_plots1, format='pdf')


def elbow_methods1(value_matrix, min_number_of_clusters,
                   max_number_of_clusters, output_plot2):
    range_k = range(int(min_number_of_clusters), int(max_number_of_clusters))
    k_m = [KMeans(n_clusters=k).fit(value_matrix) for k in range_k]
    centroids = [k.cluster_centers_ for k in k_m]
    d_k = [cdist(value_matrix, cent, 'euclidean') for cent in centroids]
    dist = [np.min(D, axis=1) for D in d_k]
    avg_within_sum_of_squares = [sum(d) / value_matrix.shape[0] for d in dist]
    fig = plt.figure()
    axes_x = fig.add_subplot(111)
    axes_x.plot(range_k, avg_within_sum_of_squares, 'b*-')
    plt.grid(True)
    plt.xlabel('Number of clusters', fontsize=15)
    plt.ylabel('Average within-cluster sum of squares', fontsize=15)
    plt.title('Elbow for KMeans clustering', fontsize=18)
    plt.savefig(output_plot2, format='pdf')


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):int(
        feature_count_end_column)]



import pandas as pd


def exclude_the_min_row_sum(feature_count_table,
                            feature_count_start_column, feature_count_end_column, min_row, output_file):
    feature_count_table_df = pd.read_table(feature_count_table)
    matrix_value = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column, feature_count_end_column)
    colum_with_gene_name = _extract_gene_matrix(feature_count_table_df)
    attribute_matrix = _extract_attributes(feature_count_table_df, feature_count_start_column)
    min_row_sum(matrix_value, attribute_matrix, colum_with_gene_name, min_row, output_file)


def _extract_value_matrix(feature_count_table_df, feature_count_start_column,
                          feature_count_end_column):
    return feature_count_table_df.iloc[:, feature_count_start_column:(
        feature_count_end_column)]


def _extract_gene_matrix(feature_count_table_df):
    gene_column = feature_count_table_df[list(filter(
        lambda col: col.startswith("Attributes"), feature_count_table_df.columns))]
    return gene_column


def _extract_attributes(feature_count_table_df,
                        feature_count_start_column):
    return feature_count_table_df.iloc[:, : feature_count_start_column]


def min_row_sum(value_matrix, attribute_matrix, gene_column, min_row, output_file):
    gene_table_final = []
    combined_df_ext = pd.concat([attribute_matrix, value_matrix], axis=1)
    summed_values = value_matrix.sum(axis=1)
    combined_df = pd.concat([gene_column, summed_values], axis=1)
    combined_df.columns = ['Attributes', 'sum_of_values']
    selected_df = combined_df[~(combined_df['sum_of_values'] <= min_row)]
    selected_df.reset_index(drop=True, inplace=True)
    my_keys = selected_df['Attributes'].tolist()
    for index, row in combined_df_ext.iterrows():
        gene = row["Attributes"]
        if gene in my_keys:
            gene_table_final.append(row)
    df_with_min_row_samples = pd.DataFrame(gene_table_final)
    df_with_min_row_samples.reset_index(drop=True, inplace=True)
    df_with_min_row_samples.to_csv(output_file, sep='\t', index=0)



import pandas as pd
from scipy.stats import spearmanr
import numpy as np


def find_complexes(tables_containing_list_complexes, protein_table,
                   feature_count_start_column, feature_count_end_column,
                   output_table):
    tables_containing_list_complexes_df = pd.read_excel(tables_containing_list_complexes)
    protein_table_df = pd.read_csv(protein_table, sep='\t')
    selected_complexes = get_into_excel_complexes_table(tables_containing_list_complexes_df,
                                                        protein_table_df,
                                                        feature_count_start_column,
                                                        feature_count_end_column)
    table_with_selected_complexes = modifying_first_letter(selected_complexes)
    table_with_complete_complexes = correlation(table_with_selected_complexes, output_table)
    finding_number_of_complexes(table_with_complete_complexes)


def get_into_excel_complexes_table(tables_containing_list_complexes, protein_table,
                                   feature_count_start_column,
                                   feature_count_end_column):
    output_df = pd.DataFrame()
    for column_idx in tables_containing_list_complexes:
        col = tables_containing_list_complexes[column_idx].dropna()
        for row in range(0, col.size):
            complex_name = col.name
            gene_name = col[row]
            selected_df = protein_table.loc[protein_table['Gene.names'] == gene_name]
            selected_df = selected_df.iloc[:, int(feature_count_start_column): feature_count_end_column]

            cols = list(selected_df)
            if selected_df.empty:
                # print("empty")
                selected_df = pd.DataFrame().append({'complex_name': complex_name, "gene": gene_name}, ignore_index=True)
            else:
                selected_df['complex_name'] = complex_name
                selected_df['gene'] = gene_name
                cols = ['complex_name', 'gene'] + cols
                selected_df = selected_df.loc[:, cols]

            output_df = output_df.append(selected_df)
    return output_df


def modifying_first_letter(table_with_complexes):
    table_capitalized = table_with_complexes.gene.str.capitalize()
    table_capitalized_last = table_capitalized.apply(lambda s: (s[:1].upper() + s[1:-1] + s[-1:].upper())[:len(s)])
    table_capitalized_df = pd.Series.to_frame(table_capitalized_last)
    table_concat = pd.concat([table_capitalized_df, table_with_complexes], axis=1)
    return table_concat


def correlation(table_with_selected_complexes, output_table):
    table_dropped = table_with_selected_complexes.iloc[:, 1:-1]
    series_by_complex = {}
    for index, row in table_dropped.iterrows():
        complex_name = row['complex_name']
        if complex_name not in series_by_complex:
            series_by_complex[complex_name] = []

        series_by_complex[complex_name].append(row.drop('complex_name'))

    table_with_selected_complexes['rho_median'] = None
    table_with_selected_complexes['rho_mean'] = None
    table_with_selected_complexes['standard_deviation'] = None
    for key in series_by_complex.keys():
        series = series_by_complex[key]
        series_df = pd.DataFrame.from_records(series).dropna(axis=1)
        if series_df.empty:
            continue

        rho, pval = spearmanr_corrected(series_df, axis=1)
        standard_deviation = series_df.stack().std()
        rho_median = calculate_median_upper_triangle(rho)
        rho_mean = calculate_mean_upper_triaglie(rho)
        table_with_selected_complexes.loc[
            table_with_selected_complexes.complex_name == key,
            'rho_median'] = rho_median
        table_with_selected_complexes.loc[
            table_with_selected_complexes.complex_name == key,
            'rho_mean'] = rho_mean
        table_with_selected_complexes.loc[
            table_with_selected_complexes.complex_name == key,
            'standard_deviation'] = \
            standard_deviation

    table_with_complete_complexes = table_with_selected_complexes.dropna(
        subset=['rho_median'], how='all')
    table_with_complete_complexes.to_csv(
        output_table, sep='\t', index=None)
    return table_with_complete_complexes


def finding_number_of_complexes(table_with_selected_complexes):
    table_with_complexes_df_grouped = table_with_selected_complexes.groupby(
        'complex_name').agg(
        lambda x: set(x)).reset_index()
    print(len(table_with_complexes_df_grouped.index))


def calculate_median_upper_triangle(matrix):
    return np.median(matrix[np.triu_indices_from(matrix, 1)])


def calculate_mean_upper_triaglie(matrix):
    return np.mean(matrix[np.triu_indices_from(matrix, 1)])


def spearmanr_corrected(a, b=None, axis=0):
    rho, pval = spearmanr(a, b, axis=axis)

    # handle edge case where rho is not a matrix (len(series_df) == 2)
    if np.isscalar(rho):
        rho = np.array([[1, rho], [rho, 1]])

    return rho, pval



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np


def d_c_graph(table_with_correlation_coefficient, percentile, index_table,
              output_plot):
    correlation_df = pd.read_table(table_with_correlation_coefficient)
    correlation_df.set_index(index_table, inplace=True)
    data = correlation_df.apply(pd.Series).unstack().reset_index(drop=True)
    data.hist()
    axes_x = data[~np.isnan(data)]
    plot = np.percentile(axes_x, percentile)
    sns.distplot(axes_x, kde=True, rug=False)
    plt.grid(False)
    plt.axvline(x=plot).set_color("black")
    plt.title("Correlation histogram "
              + '\n' +
              "The percentile value is " +
              str(plot),
              fontsize=10)
    plt.xlabel = "Correlation coefficients"

    plt.ylabel = "Number of entries"
    plt.savefig(output_plot)
    plt.show()



import pandas as pd


def move_columns_(feature_count_table, number_of_columns, output_file):
    quanti_table = pd.read_csv(feature_count_table, sep='\t')
    table_column_inverted = create_gene_quanti_table_modified(quanti_table, number_of_columns)
    table_column_inverted.to_csv(output_file, sep='\t', index=None)


def create_gene_quanti_table_modified(gene_quanti_table, number_of_columns):
    cols = gene_quanti_table.columns.tolist()
    cols = cols[-number_of_columns:] + cols[:-number_of_columns]
    gene_quanti_table_with_gene_column_modified = gene_quanti_table[cols]
    return gene_quanti_table_with_gene_column_modified





from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.metrics import silhouette_score


def silhouette_analysis(feature_count_table, feature_count_start_column,
                        feature_count_end_column,
                        min_number_of_clusters, max_number_of_clusters):
    feature_count_table_df = pd.read_table(feature_count_table)
    value_matrix = _extract_value_matrix(feature_count_table_df,
                                         feature_count_start_column, feature_count_end_column)
    silhouette_plot(value_matrix, min_number_of_clusters, max_number_of_clusters)
    score(value_matrix, min_number_of_clusters, max_number_of_clusters)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column, feature_count_end_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):feature_count_end_column]


def silhouette_plot(value_matrix, min_number_of_clusters,
                    max_number_of_clusters):
    silhouette = []
    for n_clusters in range(int(min_number_of_clusters), int(max_number_of_clusters)):
        kmeans = KMeans(n_clusters=n_clusters)
        kmeans.fit(value_matrix)
        labels = kmeans.labels_
        silhouette.append(silhouette_score(
            value_matrix, labels, metric='euclidean'))

    plt.plot(silhouette)
    plt.ylabel("Silhouette", fontsize=15)
    plt.xlabel("Number of clusters", fontsize=15)
    plt.title("Silhouette for K-means behaviour", fontsize=18)
    plt.savefig('silhouette.pdf')


def score(value_matrix, min_number_of_clusters,
          max_number_of_clusters):
    for n_clusters in range(int(min_number_of_clusters), int(max_number_of_clusters)):
        clusterer = KMeans(n_clusters=n_clusters, random_state=10)
        cluster_labels = clusterer.fit_predict(value_matrix)
        silhouette_avg = silhouette_score(value_matrix, cluster_labels)
        print("For n_clusters =", n_clusters,
              "The average silhouette_score is :", silhouette_avg)



import pandas as pd


def filtering_ercc_table(ref_feature_count_table, min_row_sum,
                         filtered_ref_feature_count_table):
    ref_feature_count_table_df = pd.read_table(ref_feature_count_table, sep='\t')
    table_filtered_ercc = create_a_new_table_filtered_ercc(ref_feature_count_table_df)
    colum_with_libraries = _extract_gene_matrix(table_filtered_ercc)
    table_filtered_min_row_sum = min_row_sum_ercc_table(table_filtered_ercc, colum_with_libraries,
                                                        min_row_sum)
    table_filtered_min_row_sum.to_csv(filtered_ref_feature_count_table, sep='\t', index=None)


def _extract_gene_matrix(feature_count_table_df):
    gene_column = feature_count_table_df[list(filter(
        lambda col: col.startswith("Libraries"), feature_count_table_df.columns))]
    return gene_column


def create_a_new_table_filtered_ercc(alignment_table):
    series = []
    for index, row in alignment_table.iterrows():
        selection = row['Libraries']
        if selection.startswith('ERCC') & ('No. of aligned reads' in selection):
            series.append(row)
    filtered_table = pd.DataFrame(series)
    return filtered_table


def min_row_sum_ercc_table(table_filtered_ercc, colum_with_libraries, min_row_sum):
    gene_table_final = []
    summed_values = table_filtered_ercc.sum(axis=1)
    combined_df = pd.concat([colum_with_libraries, summed_values], axis=1)
    combined_df.columns = ['Gene', 'sum_of_values']
    selected_df = combined_df[~(combined_df['sum_of_values'] <= min_row_sum)]
    selected_df.reset_index(drop=True, inplace=True)
    my_keys = selected_df['Gene'].tolist()
    for index, row in table_filtered_ercc.iterrows():
        gene = row["Libraries"]
        if gene in my_keys:
            gene_table_final.append(row)
    df_with_min_row_samples = pd.DataFrame(gene_table_final)
    df_with_min_row_samples.reset_index(drop=True, inplace=True)
    return df_with_min_row_samples



# coding: utf8

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='GRADitude',
    version='0.1.1',
    packages=['graditudelib'],
    author='Silvia Di Giorgio, Konrad U. Förstner',
    author1_email='digiorgio@zbmed.de',
    author2_email='foerstner@zbmed.de',
    description='A tool for the analysis of GRAD-seq data',
    url='https://github.com/foerstner-lab/GRADitude.git',
    install_requires=[
        "matplotlib == 3.1.1",
        "scipy == 1.3.1",
        "bokeh == 1.3.4",
        "Jinja2 == 2.10.3",
        "numpy == 1.17.2",
        "umap-learn == 0.3.10",
        "pytest == 5.2.1",
        "holoviews == 1.14.1",
        "pandas == 0.25.1",
        "networkx == 2.3",
        "seaborn == 0.9.0",
        "scikit-learn == 0.21.3"
    ],
    scripts=['bin/graditude'],
    license='ISC License (ISCL)',
    classifiers=[
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ]
)



import pandas as pd


def merge_features_(feature_count_table, output_file):
    feature_count_table_df = pd.read_table(feature_count_table, sep='\t')
    exctract_name_information(feature_count_table_df, output_file)


def exctract_name_information(feature_count_table_df, output_file):
    feature_count_table_df['Feature'].replace(['antisense_RNA', 'RNase_P_RNA', 'SRP_RNA', "tmRNA"],
                                              ['ncRNA', 'ncRNA', 'ncRNA', "ncRNA"], inplace=True)
    feature_count_table_df.to_csv(output_file, sep='\t', index=None)




from __future__ import print_function

from sklearn.datasets import make_blobs
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples, silhouette_score

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import pandas as pd

print(__doc__)


def main():
    table = pd.read_table('../data/k-means_Norm_MIN_ROW_SUM_100.csv_max_6cl.csv', sep="\t")
    table.set_index('Gene', inplace=True)
    value_matrix = _extract_value_matrix(table, int(11))
    silhoutte(value_matrix)


def _extract_value_matrix(feature_count_table_df,
                          feature_count_start_column):
    return feature_count_table_df.iloc[:, int(feature_count_start_column):-1]


# Generating the sample data from make_blobs
# This particular setting has one distinct cluster and 3 clusters placed close
# together.


def silhoutte(value_matrix):
    range_n_clusters = [5, 6, 7]
    for n_clusters in range_n_clusters:
        # Create a subplot with 1 row and 2 columns
        fig, (ax1, ax2) = plt.subplots(1, 2)
        fig.set_size_inches(25, 7)

        # The 1st subplot is the silhouette plot
        # The silhouette coefficient can range from -1, 1 but in this example all
        # lie within [-0.1, 1]
        ax1.set_xlim([-0.1, 1])
        # The (n_clusters+1)*10 is for inserting blank space between silhouette
        # plots of individual clusters, to demarcate them clearly.
        ax1.set_ylim([0, len(value_matrix) + (n_clusters + 1) * 10])

        # Initialize the clusterer with n_clusters value and a random generator
        # seed of 10 for reproducibility.
        clusterer = KMeans(n_clusters=n_clusters, random_state=10)
        cluster_labels = clusterer.fit_predict(value_matrix)

        # The silhouette_score gives the average value for all the samples.
        # This gives a perspective into the density and separation of the formed
        # clusters
        silhouette_avg = silhouette_score(value_matrix, cluster_labels)
        print("For n_clusters =", n_clusters,
              "The average silhouette_score is :", silhouette_avg)

        # Compute the silhouette scores for each sample
        sample_silhouette_values = silhouette_samples(value_matrix, cluster_labels)

        y_lower = 10
        for i in range(n_clusters):
            # Aggregate the silhouette scores for samples belonging to
            # cluster i, and sort them
            ith_cluster_silhouette_values = \
                sample_silhouette_values[cluster_labels == i]

            ith_cluster_silhouette_values.sort()

            size_cluster_i = ith_cluster_silhouette_values.shape[0]
            y_upper = y_lower + size_cluster_i

            color = cm.nipy_spectral(float(i) / n_clusters)
            ax1.fill_betweenx(np.arange(y_lower, y_upper),
                              0, ith_cluster_silhouette_values,
                              facecolor=color, edgecolor=color, alpha=0.7)

            # Label the silhouette plots with their cluster numbers at the middle
            ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

            # Compute the new y_lower for next plot
            y_lower = y_upper + 10  # 10 for the 0 samples

        ax1.set_title("The silhouette plot for the various clusters.")
        ax1.set_xlabel("The silhouette coefficient values")
        ax1.set_ylabel("Cluster label")

        # The vertical line for average silhouette score of all the values
        ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

        ax1.set_yticks([])  # Clear the yaxis labels / ticks
        ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

        # 2nd Plot showing the actual clusters formed
        colors = cm.nipy_spectral(cluster_labels.astype(float) / n_clusters)
        ax2.scatter(value_matrix.iloc[:, 0], value_matrix.iloc[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                    c=colors, edgecolor='k')

        # Labeling the clusters
        centers = clusterer.cluster_centers_
        # Draw white circles at cluster centers
        ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                    c="white", alpha=1, s=200, edgecolor='k')

        for i, c in enumerate(centers):
            ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                        s=50, edgecolor='k')

        ax2.set_title("The visualization of the clustered data.")
        ax2.set_xlabel("Feature space for the 1st feature")
        ax2.set_ylabel("Feature space for the 2nd feature")

        plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                      "with n_clusters = %d" % n_clusters),
                     fontsize=14, fontweight='bold')

    plt.show()


main()
